﻿#Region "FileHeader"
#If False Then

#End If '-- Options, Imports
Imports System.Reflection
Imports System.Data.Common
Imports System.ComponentModel
Imports System.Windows.Forms
Imports System.Text.RegularExpressions
Imports System.Collections.ObjectModel
#End Region 'FileHeader


Namespace System.Data

   ''' <summary>
   ''' manager-class for a dataset. associates TableAdapters with Datatables. 
   ''' TableAdapters know, how to store and fill a DataTable
   ''' </summary>
   Public Class DatasetAdapter : Inherits DatasetAdapterBase
      Public ReadOnly Provider As String
      Public ReadOnly Connection As DbConnection

      Friend ReadOnly Factory As DbProviderFactory
      Public ReadOnly Adapters As New Dictionary(Of DataTable, TableAdapter)
      Friend ReadOnly CommandBuilder As DbCommandBuilder
      Friend ReadOnly Quote, UnQuoteAll As Func(Of String, String)
      Friend ReadOnly LastRowIdSqlFunction As String     'to re-query autogenerated primKeys (smart providers only)

      Private ReadOnly _IdentityCommand As DbCommand 'to re-query autogenerated primKeys
      Private ReadOnly _FillJobs As New List(Of TableAdapter.FillJob)

      Private ReadOnly _KnownParams As DbParameterCollection    'holds a DBParam for each Column in the database
      Private ReadOnly _rgxSqlAnalyse As Regex      'analize Sql: As - keyword, Operators, Column-Identifiers, param-placeholder '?'
      'adds the param with the value to the DbParameterCollection (doing it generic is tricky)
      Private ReadOnly _AddWithValue As Func(Of DbParameterCollection, String, Object, DbParameter)
      'to detect, whether a condition expresses an inner-join
      Private Shared ReadOnly _rgxJoining As New Regex("^\W*(\w+\W*){2}join", RegexOptions.IgnoreCase)
      Public Sub New(ByVal factory As DbProviderFactory, ByVal sConnection As String, ByVal conflictOption As ConflictOption)
         Me.Factory = factory
         Connection = factory.CreateConnection
         Connection.ConnectionString = sConnection
         Provider = Connection.GetType.Namespace
         ' find out LastRowIdSqlFunction, if the provider provides a specific function to re-query autogenerated primKeys
         Dim _SmartProviders = New Dictionary(Of String, String)
         _SmartProviders.Add("System.Data.SqlClient", "SCOPE_IDENTITY")
         _SmartProviders.Add("MySql.Data.MySqlClient", "last_insert_id")
         _SmartProviders.Add("System.Data.SQLite", "last_insert_rowid")
#If False Then 'no smart providers
         _SmartProviders.Add("System.Data.OleDb", "SCOPE_IDENTITY")
         _SmartProviders.Add("System.Data.SqlServerCe", "SCOPE_IDENTITY")
#End If
         _SmartProviders.TryGetValue(Provider, LastRowIdSqlFunction)
         Using dummiCommand = factory.CreateCommand ' create a paramCollection indirectly
            _KnownParams = dummiCommand.Parameters
         End Using
         Dim addMethod = _KnownParams.GetType.GetMethod("AddWithValue")
         Me._AddWithValue = Function(params, name, value) _
            DirectCast(addMethod.Invoke(params, New Object() {name, value}), DbParameter)
         CommandBuilder = factory.CreateCommandBuilder
         With CommandBuilder
            .ConflictOption = conflictOption
            .SetAllValues = True
            If .QuotePrefix.Length = 0 Then
               .QuotePrefix = "["
               .QuoteSuffix = "]"
            End If
            Dim quoteChars = (.QuotePrefix & .QuoteSuffix).ToCharArray
            Dim fullColumn = "(?<fullColumn>[\".And("\".Between("w.*" & quoteChars), "]+)")
            Dim sPat = "(?<As>\s+As\s+)|(?<Operator>[<=>]{1,2}|\bLIKE\b|\bIN\b)|\?|" _
               & fullColumn 'pattern für gequotete und/oder volle Spaltenbezeichner anhängen
            _rgxSqlAnalyse = New Regex(sPat, RegexOptions.IgnoreCase)
            Quote = AddressOf .QuoteIdentifier
            UnQuoteAll = Function(s) String.Concat(s.Split(quoteChars))
         End With
         If LastRowIdSqlFunction.Null Then
            'Access, SqlCe schießen per Event hinter jedem Insert einen Select hinterher
            _IdentityCommand = Connection.CreateCommand
            _IdentityCommand.CommandText = "SELECT @@IDENTITY"
         End If
      End Sub

      Public Overrides Property DataSet() As DataSet
         Get
            Return _DataSet
         End Get
         Set(ByVal value As DataSet)
            If value Is _DataSet Then Return
            AddHandler value.Disposed, AddressOf _Datset_Disposed
            MyBase.DataSet = value
            Adapters.Clear()
            Using Connection.EnterOpen
               Dim cnt = New Counter
               For Each tb As DataTable In _DataSet.GetRankedTables
                  Adapters(tb) = New TableAdapter(tb, Me, cnt.Down)
               Next
            End Using
         End Set
      End Property

      Friend Sub AddKnownParam(ByVal table As DataTable, ByVal param As DbParameter)
         Dim fullName = String.Concat(table.TableName, ".", param.SourceColumn)
         If _KnownParams.Contains(fullName) Then Return
         param = param.Clone
         param.ParameterName = fullName
         _KnownParams.Add(param)
      End Sub

      Friend Sub AddFillJobs(ByVal jobs As IEnumerable(Of TableAdapter.FillJob))
         '_FillJobs is a kinda consumer-Heap: Inserting first job will start executing. Execution may cause adding of more jobs (to fill childTables), wich will be executed too - respecting their Rank.
         With _FillJobs
            Dim start = .Count = 0
            For Each job In jobs
               Dim i = .BinarySearch(job)
               If i < 0 Then
                  .Insert(i Xor -1, job)
               Else
                  .Item(i) = job ' last in wins
               End If
            Next
            If Not start Then Return
            Using DataSet.Connection.EnterOpen
               While .Count > 0
                  .Pop.Execute()
               End While
            End Using
         End With
      End Sub

      Private Function ParamFromKnownParams(ByVal table As DataTable, ByVal tryal As String, ByVal aliases As Dictionary(Of String, String)) As DbParameter
         Dim splits = UnQuoteAll(tryal).Split("."c)
         If splits.Length < 2 Then splits = New String() {table.TableName, splits(0)}
         tryal = aliases.GetOrDefault(splits(0), splits(0)).And(".", splits(1))
         Return If(_KnownParams.Contains(tryal), _KnownParams(tryal).Clone, Nothing)
      End Function

      Friend Function CreateSelectCommand(ByVal table As DataTable, preambles As String, ByVal conditions As String, ByVal paramValues As IEnumerable) As DbCommand
         Dim sTable = table.TableName
         Dim oValues = paramValues.GetEnumerator
         Dim selector = "SELECT ".And(preambles, " ", Quote(sTable), ".* FROM ")
         If Not _rgxJoining.IsMatch(conditions) Then
            'conditions ist keine Join-Anweisung -> TabellenNamen an den selector hängen
            selector &= Quote(sTable) & " "
         End If
         Dim cmd = Connection.CreateCommand
         'dummi-matches zufügen, denn die Suche greift 2 zurück oder 1 vor
         Dim mts = _rgxSqlAnalyse.Matches("d d ".And(conditions, " d"))
         Dim aliases As New Dictionary(Of String, String)
         Dim getColkey = Function(i As Integer) UnQuoteAll(mts(i).Value)
         Dim cnt = New Counter
         For i = 2 To mts.Count - 1
            ' perform 2 concerns: a) store mapping for Sql-Aliases, b) configure DBParameter with value,
            '°when paramPlaceholder '?' occurs in conditions
            Dim mtsCurr = mts(i)
            If mtsCurr.Groups("As").Success Then ' Umbenennung merken
               aliases.Add(getColkey(i + 1), getColkey(i - 1))
            ElseIf mtsCurr.Value = "?" Then ' configure DBParameter
               Dim before = mts(i - 1)
               Dim colKey = ""
               If before.Groups("Operator").Success Then
                  colKey = getColkey(i - 2) 'column on the left of '?'
               ElseIf before.Value = "?" Then 'IN-Operator: previous match also was '?'
                  'take the previous parameterName as colKey
                  colKey = cmd.Parameters(cmd.Parameters.Count - 1).SourceColumn
               ElseIf mts(i + 1).Groups("Operator").Success Then
                  colKey = getColkey(i + 2) 'column on the right of '?'
               End If
               Dim param = ParamFromKnownParams(table, colKey, aliases)
               If Not oValues.MoveNext Then Throw New ArgumentException( _
                     "conditions contains more placeholders than paramValues are passed")
               If param.Null Then
                  'unknown Sql-Function: hopefully AddWithValue will generate valid Param
                  _AddWithValue(cmd.Parameters, "@p" & cnt.Up, oValues.Current)
               Else
                  param.ParameterName = "@p" & cnt.Up
                  param.Value = oValues.Current
                  cmd.Parameters.Add(param)
               End If
            End If
         Next
         If LastRowIdSqlFunction.NotNull Then
            'smart DBProvider require parameter-names - only '?' doesn't work. 
            Dim splits = conditions.Split("?"c)
            For i = 0 To cmd.Parameters.Count - 1
               splits(i) &= cmd.Parameters(i).ParameterName
            Next
            conditions = String.Concat(splits)
         End If
         cmd.CommandText = selector & conditions
         Return cmd
      End Function

      Protected Overrides Sub _Save(ByVal attaches As List(Of DataTableAttach))
         Dim tas = attaches.Select(Function(att) Adapters(att.Table)).ToArray
         Dim dt = Date.Now
         Dim con = DataSet.Connection
         Using x = con.EnterOpen, tr = con.BeginTransaction()
            tas.ForEach(Sub(ta) ta.PrepareChangedRows())
            Dim suspendeds = SuspendInserterBindingSources(attaches)
            For Each ta In tas
               ta.AllCommands.Skip(1).ForEach(Sub(cmd) cmd.Transaction = tr)
               ta.UpdateByRowstate(DataRowState.Added Or DataRowState.Modified)
            Next
            suspendeds.ForEach(Sub(att) att.SuspendLogic(False, DataTableAttach.Suspend.BindingSources))
            Dim cascadeAcceptRules = New _AcceptruleCascader(DataSet)
            tas.ForEach(Sub(ta) ta.UpdateByRowstate(DataRowState.Deleted))   'verbleibende Deletes mit AcceptRule.Cascade
            cascadeAcceptRules.Restore()
            tr.Commit()
         End Using
         Dbg("Update-Duration:", (Date.Now - dt).TotalMilliseconds, "ms")
      End Sub

      Private Function SuspendInserterBindingSources(attaches As List(Of DataTableAttach)) As List(Of DataTableAttach)
         'BindingSources der Child-Tables von geaddeten Datensätzen müssen disabled werden, denn wenn die UpdateRule den Foreignkey updatet tritt kurzzeitig ein ungültiger Zustand in angebundenen DatagridViewComboboxColumns auf.
         'Hier werden kunstreich und effizient genau die Tabellen gefiltert, die Rows enthalten, deren ForeignKey sich durch Id-Requery einer inserteten ParentRow ändern würde
         Dim lst = New List(Of DataTableAttach)
         Dim rlsGroups = From att In attaches
                         From rl In att.Table.ParentRelations.Cast(Of DataRelation)()
                         Where rl.ParentColumns(0).AutoIncrement AndAlso rl.ChildKeyConstraint.NotNull _
                         AndAlso Adapters(rl.ParentTable).ChangedRows(DataRowState.Added).NotNull
                         Select rl Group By rl.ChildTable Into Group Select DirectCast(Group, DataRelation())
         For Each grp In rlsGroups
            Dim ta = Adapters(grp(0).ChildTable)
            For Each state In {DataRowState.Added, DataRowState.Modified}
               Dim rows = ta.ChangedRows(state)
               If rows.NotNull AndAlso grp.Any(Function(rl) rows.Any(Function(rw) rw.GetParentRow(rl).RowState = DataRowState.Added)) Then
                  Dim att = ta.Table.Attach
                  att.SuspendLogic(True, DataTableAttach.Suspend.BindingSources)
                  lst.Add(att)
                  Exit For
               End If
            Next
         Next
         Return lst
      End Function

      Private Class _AcceptruleCascader : Inherits List(Of Tuple(Of ForeignKeyConstraint, AcceptRejectRule))
         Public Sub New(ByVal dts As DataSet)
            For Each rl As DataRelation In dts.Relations
               Dim ck = rl.ChildKeyConstraint
               If ck.Null Then Continue For
               MyBase.Add(Tuple.Create(ck, ck.AcceptRejectRule))
               ck.AcceptRejectRule = AcceptRejectRule.Cascade
            Next
         End Sub
         Public Sub Restore()
            For Each tpl In Me
               tpl.Item1.AcceptRejectRule = tpl.Item2
            Next
         End Sub
      End Class

      Public Sub RequeryID(ByVal sender As Object, ByVal e As RowUpdatedEventArgs)
         If e.StatementType <> StatementType.Insert Then Return
         _IdentityCommand.Transaction = DirectCast(e.Command.Transaction, DbTransaction)
         Dim primCol = e.Row.Table.PrimaryKey(0)
         Dim primVal = _IdentityCommand.ExecuteScalar
         e.Row(primCol) = Convert.ChangeType(primVal, primCol.DataType)
      End Sub

      Friend IdRequerier As Action(Of DatasetAdapter, DbDataAdapter)
      Public Sub InitRequery_ID(ByVal idRequerier As Action(Of DatasetAdapter, DbDataAdapter))
         Me.IdRequerier = idRequerier
      End Sub

      Private Sub _Datset_Disposed(ByVal s As Object, ByVal e As EventArgs)
         For Each d In New IDisposable() { _
               Connection, CommandBuilder, _IdentityCommand}
            If d.NotNull Then d.Dispose()
         Next
         For Each d In Adapters.Keys : d.Dispose() : Next
      End Sub

      Protected Overrides Sub _Fill(ByVal tables As Collections.Generic.IEnumerable(Of DataTable))
         If Not tables.Any Then tables = _DataSet.Tables.Cast(Of DataTable)()
         Using _DatasetAttach.BeginSuspendAll
            AddFillJobs(From tb In tables Select New TableAdapter.FillJob(Adapters(tb)))
         End Using
      End Sub
   End Class

End Namespace