
Imports System.Collections

Namespace System.Collections.Generic

   ''' <summary>
   ''' abstrakte Basisklasse, die die Schnittstelle IEnumerable(Of T) vererbt, und so deren Implementation vereinfacht
   ''' </summary>
   Public MustInherit Class Enumerable(Of T) : Implements IEnumerable(Of T), IEnumerator(Of T)
      Private _IsPrepared As Boolean
      Private _Current As T = Nothing

      ''' <summary> fordert vom Erben das nchste Element der Enumeration an </summary> 
      ''' <returns> True: wenn ein Element gegeben wurde, False: wenn die Enumeration beendet ist </returns>
      Protected MustOverride Function TryMovenext(ByRef Current As T) As Boolean

      ''' <summary> ein Erbe mu entweder Prepare() oder CleanUp() berschreiben,  um seinen Ausgangszustand wieder herzustellen </summary>
      Protected Overridable Sub Prepare()
      End Sub
      ''' <summary> ein Erbe mu entweder Prepare() oder CleanUp() berschreiben,  um seinen Ausgangszustand wieder herzustellen </summary>
      Protected Overridable Sub CleanUp()
         _IsPrepared = False
      End Sub
      Public Function GetEnumerator() As IEnumerator(Of T) Implements IEnumerable(Of T).GetEnumerator
         If _IsPrepared Then Throw New Exception(String.Concat( _
                  "IEnumerable(Of ", GetType(T).Name, ") ", _
                  "kann nicht mehrere Enumeratoren gleichzeitig bereitstellen."))
         Prepare()
         _IsPrepared = True
         Return Me
      End Function
      Private Function GetEnumerator1() As IEnumerator Implements IEnumerable.GetEnumerator
         Return GetEnumerator()
      End Function
      Private ReadOnly Property Current() As T Implements IEnumerator(Of T).Current
         Get
            Return _Current
         End Get
      End Property
      Private ReadOnly Property Current1() As Object Implements IEnumerator.Current
         Get
            Return _Current
         End Get
      End Property
      <DebuggerStepThrough()> _
      Private Function MoveNext() As Boolean Implements IEnumerator.MoveNext
         Return TryMovenext(_Current)
      End Function
      Private Sub Reset() Implements IEnumerator.Reset
         'habe noch nie mitbekommen, da dieser Schnittstellenmember aufgerufen wurde
         Prepare()
      End Sub
      Private Sub Dispose() Implements IDisposable.Dispose
         CleanUp()
      End Sub
   End Class 'Enumerable(Of T)

End Namespace

'Imports System.Collections
'
'Namespace System.Collections.Generic
'
'   ''' <summary>
'   ''' abstrakte Basisklasse, die die Schnittstelle IEnumerable(Of T) vererbt, und so deren 
'   ''' Implementation vereinfacht
'   ''' </summary>
'   Public MustInherit Class Enumerable(Of T) : Implements IEnumerable(Of T)
'
'      Private _enrt As New Enumerator(Me)
'      Private _IsPrepared As Boolean
'
'      ''' <summary>
'      ''' fordert vom Erben das nchste Element der Enumeration an
'      ''' </summary>
'      ''' <returns>
'      ''' True: wenn ein Element gegeben wurde, False: wenn die Enumeration beendet ist
'      ''' </returns>
'      Protected MustOverride Function TryMovenext(ByRef Current As T) As Boolean
'
'      ''' <summary>
'      ''' ein Enumerable(Of T) - Erbe mu entweder Prepare() oder CleanUp() berschreiben, 
'      ''' um seinen Ausgangszustand wieder herzustellen
'      ''' </summary>
'      Protected Overridable Sub Prepare()
'      End Sub
'      ''' <summary>
'      ''' ein Enumerable(Of T) - Erbe mu entweder Prepare() oder CleanUp() berschreiben, 
'      ''' um seinen Ausgangszustand wieder herzustellen
'      ''' </summary>
'      Protected Overridable Sub CleanUp()
'         _IsPrepared = False
'      End Sub
'
'      Public Function GetEnumerator() As IEnumerator(Of T) Implements IEnumerable(Of T).GetEnumerator
'         If _IsPrepared Then Throw New Exception(String.Concat( _
'                  "IEnumerable(Of ", GetType(T).Name, ") ", _
'                  "kann nicht mehrere Enumeratoren gleichzeitig bereitstellen."))
'         Prepare()
'         _IsPrepared = True
'         Return _enrt
'      End Function
'
'      Private Function GetEnumerator1() As IEnumerator Implements IEnumerable.GetEnumerator
'         Return GetEnumerator()
'      End Function
'
'      '--------- nested Class: Enumerable(Of T).Enumerator -----------
'      'Der Enumerator macht nicht viel: Er leitet nur Aufrufe an die Schnittstelle um
'      Public Class Enumerator : Implements IEnumerator(Of T)
'         Private _Current As T = Nothing
'
'         Private _enmbl As Enumerable(Of T)
'
'         Public Sub New(enmbl As Enumerable(Of T))
'            _enmbl = enmbl
'         End Sub
'
'         Private ReadOnly Property Current() As T Implements IEnumerator(Of T).Current
'            Get
'               Return _Current
'            End Get
'         End Property
'
'         Private ReadOnly Property Current1() As Object Implements IEnumerator.Current
'            Get
'               Return _Current
'            End Get
'         End Property
'
'         <DebuggerStepThrough()> _
'         Private Function MoveNext() As Boolean Implements IEnumerator.MoveNext
'            Return _enmbl.TryMovenext(_Current)
'         End Function
'
'         Private Sub Prepare() Implements IEnumerator.Reset
'            'habe noch nie mitbekommen, da dieser Schnittstellenmember aufgerufen wurde
'            _enmbl.Prepare()
'         End Sub
'
'         Private Sub Dispose() Implements IDisposable.Dispose
'            _enmbl.CleanUp()
'         End Sub
'
'      End Class 'Enumerable(Of T).Enumerator
'
'   End Class 'Enumerable(Of T)
'
'
'End Namespace
