﻿Imports System.Runtime.CompilerServices
Imports System.ComponentModel
Imports System.Data.Common
Imports System.IO
Imports System.Text

Namespace System.Data

   Public Module DatasetX


      <Extension()> _
      Public Sub EnsureExistUnique(Of T As DataRow)(ByRef row As T, ParamArray keyCols() As DataColumn)
         Dim tbl = DirectCast(row.Table, TypedTableBase(Of T))
         Dim rw = row
         If keyCols.Length = 0 Then keyCols = tbl.PrimaryKey
         Dim keyVals = (From col In keyCols Select col, val = rw(col)).ToArray
         rw = tbl.All.FirstOrDefault(Function(rw2) keyVals.All(Function(pv) rw2(pv.col).Equals(pv.val)))
         If rw.Null Then
            tbl.Rows.Add(row)
            If keyCols.Any(Function(clm) clm.AutoIncrement) Then row.AcceptChanges()
         Else
            row.Be(rw)
         End If
      End Sub

      <Extension()> _
      Public Sub CopyTo(Of T As DataRow)(subj As T, other As T)
         Dim itms = subj.ItemArray
         Dim tb = other.Table
         For Each indx In tb.PrimaryKey.Select(Function(cl) cl.Ordinal)
            itms(indx) = other(indx)
         Next
         other.ItemArray = itms
      End Sub

      <Extension()> _
      Public Function All(Of T As DataRow)(subj As TypedTableBase(Of T), Optional filter As Func(Of T, Boolean) = Nothing) As IEnumerable(Of T)
         All = DirectCast(subj, DataTable).All.Cast(Of T)()
         If filter.NotNull Then All = All.Where(filter)
      End Function
      <Extension()> _
      Public Function All(subj As DataTable) As IEnumerable(Of DataRow)
         Return From rw In subj Where 0 = (rw.RowState And (DataRowState.Deleted Or DataRowState.Detached))
      End Function

      <Extension()> _
      Public Function ChildRels(subj As DataTable) As IEnumerable(Of DataRelation)
         Return subj.ChildRelations.Cast(Of DataRelation)()
      End Function

      <Extension()> _
      Public Function ParentRels(subj As DataTable) As IEnumerable(Of DataRelation)
         Return subj.ParentRelations.Cast(Of DataRelation)()
      End Function

      <Extension()> _
      Public Function ChildTables(subj As DataTable) As IEnumerable(Of DataTable)
         Return From rl In subj.ChildRelations.Cast(Of DataRelation)() _
            Where rl.ChildKeyConstraint.NotNull Select rl.ChildTable
      End Function

      <Extension()> _
      Public Function ParentTables(subj As DataTable) As IEnumerable(Of DataTable)
         Return subj.ParentRelations.Cast(Of DataRelation).Select(Function(rl) rl.ParentTable)
      End Function

      Private _ClearingTables As New HashSet(Of DataTable)
      <Extension()> _
      Public Sub ClearRecursive(subj As DataTable)
         If subj.Rows.Count = 0 OrElse Not _ClearingTables.Add(subj) Then Return
         subj.ChildTables.ForEach(AddressOf ClearRecursive)
         _ClearingTables.Remove(subj)
         subj.Clear()
      End Sub

      ''' <summary>sort tables, so that no childtable is before its parentTable</summary>
      <Extension()> _
      Public Function GetRankedTables(dts As DataSet) As List(Of DataTable)
         Return New TopologicSort(dts.Tables.Cast(Of DataTable)).rankedTables
      End Function
      <Extension()> _
      Public Function GetRankedTables(subj As IEnumerable(Of DataTable)) As List(Of DataTable)
         Return New TopologicSort(subj).rankedTables
      End Function
      Private Class TopologicSort
         'Bevor ein Node aus dem Hashset in die Ergebnisliste verschoben wird,
         ' werden rekursiv all seine Vorgänger verschoben
         Public rankedTables As List(Of DataTable)
         Public tableSrc As HashSet(Of DataTable)
         Public Sub New(dts As IEnumerable(Of DataTable))
            tableSrc = New HashSet(Of DataTable)(dts)
            rankedTables = New List(Of DataTable)(tableSrc.Count)
            While tableSrc.Count > 0
               RecurseRankedTables(tableSrc.First)
            End While
         End Sub
         Public Sub RecurseRankedTables(tb As DataTable)
            If Not tableSrc.Remove(tb) Then Return 'gegen Kreisbezüge sichern
            tb.ParentTables.ForEach(AddressOf RecurseRankedTables)
            rankedTables.Add(tb)
         End Sub
      End Class

   End Module

End Namespace
