﻿Imports System.Text.RegularExpressions

#If False Then
Encoding.Default            'Ansi (CP1252 bei deutschem Windows)
Encoding.GetEncoding(1252)  'Ansi
Encoding.UTF8               'UTF8
Encoding.GetEncoding(65001) 'UTF8
Encoding.Unicode            'UTF16LE
Encoding.GetEncoding(437)   'OEM (DOS) 437
Encoding.ASCII              'Nur 7-Bit Zeichen (bis Zeichen 127)
#End If

Imports System.Runtime.CompilerServices
Imports System.Text
Imports System.Xml.Serialization

Namespace System.IO

   Public Module SystemIOX

      Private _PathReplacer As MatchEvaluator = Function(mt As Match) IO.Path.GetFullPath(mt.Groups(1).Value)
      Private _QuotedReplacer As MatchEvaluator = Function(mt As Match) """".And(Path.GetFullPath(mt.Groups(1).Value), """")
      Private Function CreateQuotedDirPattern() As Regex
         Dim escapeds = From c In Path.GetInvalidPathChars Select "\x" & Convert.ToInt32(c).ToString("x2")
         Return New Regex("""?((\.{2}\\)+[^;".And(String.Concat(escapeds), "]+)""?"))
      End Function
      <Extension()> _
      Public Function QuotedFullnamed(ByVal s As String) As String
         Return CreateQuotedDirPattern.Replace(s, _QuotedReplacer)
      End Function
      <Extension()> _
      Public Function Fullnamed(ByVal s As String) As String
         Return CreateQuotedDirPattern.Replace(s, _PathReplacer)
      End Function

      <Extension(), DebuggerStepThrough()> _
      Public Function Combine(di As DirectoryInfo, ParamArray segments As String()) As String
         Dim strings(segments.Length) As String
         strings(0) = di.FullName
         Array.Copy(segments, 0, strings, 1, segments.Length)
         Return String.Join(Path.DirectorySeparatorChar, Array.ConvertAll(strings, Function(s) s.Trim(Path.DirectorySeparatorChar)))
      End Function

      <Extension(), DebuggerStepThrough()> _
      Public Function GetFullnameWithoutExtension(fi As FileInfo) As String
         Return GetFullnameWithoutExtension(fi.FullName)
      End Function

      Public Function GetFullnameWithoutExtension(s As String) As String
         Dim l = Path.GetExtension(s).Length
         If Not s.Contains(":") Then s = Path.GetFullPath(s)
         Return s.Substring(0, s.Length - l)
      End Function


      ''' <summary> kopiert von einem Stream in einen anderen </summary>
      ''' <remarks> 
      ''' Es gibt Streams ohne festgelegtes Ende (zB. NetworkStream).
      ''' In solchem Fall **muß** 'count' angegeben werden.
      ''' </remarks>
      <Extension()> _
      Public Sub WriteTo( _
                 readStream As Stream, _
                 writeStream As Stream, _
                Optional count As Long = -1, _
                Optional bufSize As Integer = 1024)
         Dim buf(bufSize - 1) As Byte
         If count < 0 AndAlso readStream.CanSeek Then
            count = readStream.Length - readStream.Position
         End If
         If count < 0 Then
            ' Durch 0-Byte-Lesevorgang terminierte Kopier-Schleife
            ' Ein NetworkStream würde ein Timeout-Problem verursachen
            Do
               Dim portion = readStream.Read(buf, 0, bufSize)
               If portion = 0 Then Return
               writeStream.Write(buf, 0, portion)
            Loop
         Else
            ' zähler-gesteuerte Kopier-Schleife
            Do
               If count < bufSize Then bufSize = CInt(count)
               Dim portion = readStream.Read(buf, 0, bufSize)
               If portion = 0 Then Throw New ArgumentException( _
                  "Die angegebene Anzahl Bytes konnte nicht aus dem Lese-Stream gelesen werden", _
                  "readStream + count")
               count -= portion
               writeStream.Write(buf, 0, portion)
            Loop Until count = 0
         End If
      End Sub

      ''' <summary>Achtung! FileInfo.OpenWrite ist ungeeignet, denn es erneuert nicht unbedingt die ganze Datei</summary>
      <Extension()> _
      Public Sub SerializeXml(fi As FileInfo, obj As Object)
         Using strm = fi.Open(FileMode.Create)
            strm.SerializeXml(obj)
         End Using
      End Sub
      <Extension()> _
      Public Sub SerializeXml(strm As Stream, obj As Object)
         Call (New XmlSerializer(obj.GetType)).Serialize(strm, obj)
      End Sub
      <Extension()> _
      Public Function DeserializeXml(Of T)(fi As FileInfo) As T
         Using strm = fi.OpenRead
            Return strm.DeserializeXml(Of T)()
         End Using
      End Function
      <Extension()> _
      Public Function DeserializeXml(Of T)(strm As Stream) As T
         Return DirectCast((New XmlSerializer(GetType(T))).Deserialize(strm), T)
      End Function

      ''' <summary>Achtung! FileInfo.OpenWrite ist ungeeignet, denn es erneuert nicht unbedingt die ganze Datei</summary>
      <Extension()> _
      Public Sub Serialize(fi As FileInfo, obj As Object)
         Using strm = fi.Open(FileMode.Create)
            strm.Serialize(obj)
         End Using
      End Sub
      <Extension()> _
      Public Sub Serialize(strm As Stream, obj As Object)
         Call (New Runtime.Serialization.Formatters.Binary.BinaryFormatter).Serialize(strm, obj)
      End Sub
      <Extension()> _
      Public Function Deserialize(Of T)(fi As FileInfo) As T
         Using strm = fi.OpenRead
            Return strm.Deserialize(Of T)()
         End Using
      End Function
      <Extension()> _
      Public Function Deserialize(Of T)(strm As Stream) As T
         Return DirectCast((New Runtime.Serialization.Formatters.Binary.BinaryFormatter).Deserialize(strm), T)
      End Function

      <Extension()> _
      Public Function ReadAllLines(rf As FileInfo) As List(Of String)
         Dim Lst As New List(Of String)
         Using sr = New StreamReader(rf.FullName)
            Do Until sr.EndOfStream
               Lst.Add(sr.ReadLine)
            Loop
         End Using
         Return Lst
      End Function
   End Module

End Namespace